"""
Sarmat.
Описание сущностей.
Базовый класс для описания моделей.
"""
from dataclasses import asdict, dataclass, fields
from typing import Any

from sarmat.core.constants import DurationType, IntervalType


@dataclass
class BaseModel:

    @property
    def sarmat_fields(self):
        return [fld.name for fld in fields(self)]

    @property
    def as_dict(self):
        return asdict(self)


@dataclass
class BaseIdModel:

    id: int = 0


@dataclass
class BaseUidModel:

    uid: str = ""


@dataclass
class BaseCatalogModel:
    """Базовая модель для описания справочников"""

    cypher: str     # шифр (константа)
    name: str       # название


@dataclass
class CustomAttributesModel:

    custom_attributes: dict[str, Any] | None = None

    @property
    def custom_fields(self) -> list[str]:
        return list(self.custom_attributes.keys()) if self.custom_attributes else []


@dataclass
class PersonModel(BaseModel):
    """Данные человека"""

    last_name: str      # фамилия
    first_name: str     # имя
    middle_name: str    # отчество
    male: bool          # пол: М


@dataclass
class DurationItemModel(BaseModel):
    """Базовый элемент для описания продолжительности"""

    duration_type: DurationType     # тип продолжительности
    value: int                      # значение
    in_activity: bool               # признак активной фазы
    position: int = 0               # номер в последовательности
    id: int = 0                     # идентификатор записи


@dataclass
class DurationModel(BaseIdModel, CustomAttributesModel, BaseCatalogModel):
    """Составная модель с описанием продолжительности"""

    values: list[DurationItemModel] | None = None    # последовательность из отрезков времени


@dataclass
class IntervalItemModel(BaseModel):
    """Базовая модель для описания интервала"""

    interval_type: IntervalType     # тип интервала
    values: list[int]               # список значений
    in_activity: bool               # признак активной фазы
    position: int = 0               # номер в последовательности
    id: int = 0                     # идентификатор записи


@dataclass
class IntervalModel(BaseIdModel, CustomAttributesModel, BaseCatalogModel):
    """Составная модель с описанием интервала"""

    values: list[IntervalItemModel] | None = None     # описание сложного интервала



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Вычисления с участием Sarmat объектов.
"""

__all__ = (
    "calculate_route_metrics", "get_timedelta_from_duration", "get_timedelta_from_duration_item", "get_month_increase",
)

from .duration import (
    get_month_increase,
    get_timedelta_from_duration,
    get_timedelta_from_duration_item,
)
from .metrics import calculate_route_metrics



"""
Sarmat.
Ядро пакета.
Используемые константы.
"""
__all__ = (
    "RoadType", "IntervalType", "LocationType", "SettlementType", "StationType", "JourneyType", "JourneyClass",
    "JourneyState", "VehicleType", "CrewType", "PermitType", "ErrorClass", "MessageType", "SarmatException",
    "SarmatExpectedAttributeError", "SarmatNotFilledAttribute", "SarmatWrongOperationError", "SarmatWrongTypeAttribute",
    "SarmatWrongValueError", "PlaceKind", "PlaceType", "PlaceState", "DATE_FORMAT", "DATETIME_FORMAT",
    "FULL_TIME_FORMAT", "FULL_DATETIME_FORMAT", "TIME_FORMAT", "RouteType", "DurationType", "DurationMonthCalcStrategy",
    "month_len",
)

from .exception_constants import (
    ErrorClass,
    MessageType,
    SarmatException,
    SarmatExpectedAttributeError,
    SarmatNotFilledAttribute,
    SarmatWrongOperationError,
    SarmatWrongTypeAttribute,
    SarmatWrongValueError,
)
from .sarmat_constants import (
    CrewType,
    DurationMonthCalcStrategy,
    DurationType,
    IntervalType,
    LocationType,
    PermitType,
    PlaceKind,
    PlaceType,
    PlaceState,
    RoadType,
    RouteType,
    SettlementType,
    StationType,
    JourneyType,
    JourneyClass,
    JourneyState,
    VehicleType,
    month_len,
)

from .formats import (
    DATE_FORMAT,
    DATETIME_FORMAT,
    FULL_DATETIME_FORMAT,
    FULL_TIME_FORMAT,
    TIME_FORMAT,
)



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Вычисления с участием Sarmat объектов.
Вычисления продолжительности.
"""
from datetime import (
    date,
    timedelta,
)

from sarmat.core.constants import (
    DurationMonthCalcStrategy,
    DurationType,
    month_len,
)
from sarmat.core.context.models import (
    DurationItemModel,
    DurationModel,
)

from .const import (
    max_month_len,
    months_in_year,
)


def get_timedelta_from_duration(
    duration: DurationModel,
    calculation_date: date,
    calc_strategy: DurationMonthCalcStrategy,
    only_active: bool = False,
) -> timedelta:
    """Вычисление приращения по времени на основе объекта продолжительности.

    Args:
        duration: объект продолжительности
        calculation_date: дата на момент вычисления
        calc_strategy: стратегия обработки продолжительности в месяц
        only_active: признак учёта только времени активности

    Returns: приращение по времени

    """
    time_delta = timedelta(minutes=0)
    start_date = calculation_date + time_delta

    if duration.values:
        for val in duration.values:
            if only_active and not val.in_activity:
                continue
            time_delta += get_timedelta_from_duration_item(val, start_date, calc_strategy)
            start_date = calculation_date + time_delta

    return time_delta


def get_timedelta_from_duration_item(
    duration_item: DurationItemModel,
    calculation_date: date,
    calc_strategy: DurationMonthCalcStrategy,
) -> timedelta:
    """Преобразование элемента продолжительности в приращение по времени."""

    if duration_item.duration_type == DurationType.MONTH:
        args = {
            DurationType.DAY.as_cypher: get_month_increase(duration_item.value, calculation_date, calc_strategy),
        }
    elif duration_item.duration_type == DurationType.YEAR:
        args = {
            DurationType.DAY.as_cypher: get_month_increase(duration_item.value*12, calculation_date, calc_strategy),
        }
    else:
        args = {duration_item.duration_type.as_cypher: duration_item.value}
    return timedelta(**args)


def get_month_increase(
    value: int,
    calculation_date: date,
    calc_strategy: DurationMonthCalcStrategy,
) -> int:
    """Вычисление приращения по дням для месяцев."""
    current_month = calculation_date.month
    current_day = calculation_date.day
    current_year = calculation_date.year
    current_month_len, target_month_len = 0, 0

    # составление годовых кругов по количеству затронутых месяцев
    month_cycle = []
    months_count = current_month + value
    years_count, mod = divmod(months_count, months_in_year)
    if mod:
        years_count += 1

    # учитываем количество дней в високосных годах
    feb_pos = 1
    for i in range(years_count):
        month_cycle.extend(month_len)
        year = current_year + i
        if not year % 4:
            month_cycle[feb_pos] += 1
        feb_pos += months_in_year
    month_cycle_gen = (m for m in month_cycle)

    for _ in range(current_month):
        current_month_len = next(month_cycle_gen)
    # количество дней до конца текущего месяца
    days_increase = current_month_len - current_day

    if value > 1:
        if calc_strategy == DurationMonthCalcStrategy.FULL:
            for _ in range(value - 1):
                days_increase += max_month_len
                next(month_cycle_gen)
        else:
            for _ in range(value - 1):
                days_increase += next(month_cycle_gen)

    target_month_len = next(month_cycle_gen)
    # если текущая дата не превышает максимальную дату месяца,
    # то вычисляем количество дней как число дней до текущего месяца + текущая дата
    if current_day <= target_month_len:
        days_increase += current_day
    else:
        # в противном случае применяем стратегию вычисления дней для конца месяца
        # Откат к наименьшему дню месяца
        if calc_strategy == DurationMonthCalcStrategy.DOWN:
            days_increase += target_month_len
        elif calc_strategy == DurationMonthCalcStrategy.MOVE:
            days_increase += target_month_len + (current_day - target_month_len)
        elif calc_strategy == DurationMonthCalcStrategy.FULL:
            days_increase += max_month_len

    return days_increase



"""
Sarmat.
Ядро пакета.
Константы.
Константы бизнес логики.
"""
from collections import namedtuple
from enum import Enum
from typing import Any


month_len = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
SarmatConstant = namedtuple("SarmatConstant", "id cypher name")


class SarmatAttribute(SarmatConstant, Enum):
    """Встроенные значения атрибутов в Sarmat."""

    @property
    def as_text(self) -> str:
        """Описание значения."""
        val: SarmatConstant = self.value
        return val.name

    @property
    def as_cypher(self) -> str:
        """Получение строковой константы."""
        val: SarmatConstant = self.value
        return val.cypher

    @property
    def id(self) -> Any:
        val: SarmatConstant = self.value
        return val.id


class RoadType(SarmatAttribute):
    """Тип дорожного покрытия."""
    PAVED = 1, "", "Дорога с твердым покрытием"
    DIRT = 2, "", "Грунтовая дорога"
    HIGHWAY = 3, "", "Магистраль"
    TOLL_ROAD = 4, "", "Платная дорога"


class DurationType(SarmatAttribute):
    """Тип продолжительности."""
    MINUTE = 1, "minutes", "Минуты"
    HOUR = 2, "hours", "Часы"
    DAY = 3, "days", "Дни"
    WEEK = 4, "weeks", "Недели"
    MONTH = 5, "months", "Месяцы"
    YEAR = 6, "years", "Годы"


class IntervalType(SarmatAttribute):
    """Тип интервала."""
    MINUTE = 1, "minute", "По минутам"
    HOUR = 2, "hour", "По часам"
    DAY = 3, "day", "По дням"
    WEEK = 4, "week", "По неделям"
    MONTH = 5, "month", "По месяцам"
    YEAR = 6, "year", "По годам"
    EVEN = 7, "even", "По четным дням месяца"
    ODD = 8, "odd", "По нечетным дням месяца"
    DAYS = 9, "days", "По числам месяца"
    DAYS_OF_WEEK = 10, "dow", "По дням недели"


class LocationType(SarmatAttribute):
    """Тип территориального образования."""
    COUNTRY = 1, "", "Страна"
    DISTRICT = 2, "респ.", "Республика"
    REGION = 3, "кр.", "Край"
    PROVINCE = 4, "обл.", "Область"
    AREA = 5, "р-н", "Район"


class SettlementType(SarmatAttribute):
    """Тип населенного пункта."""
    CITY = 1, "г.", "Город"
    SETTLEMENT = 2, "пос.", "Поселок"
    TOWNSHIP = 3, "с.", "Село"
    HAMLET = 4, "дер.", "Деревня"
    COUNTRYSIDE = 5, "ст.", "Станица"
    FARM = 6, "х.", "Хутор"
    VILLAGE = 7, "сл.", "Слобода"
    TURN = 8, "пов.", "Поворот"
    POINT = 9, "м.", "Место"


class StationType(SarmatAttribute):
    """Типы станций."""
    STATION = 1, "АВ", "Автовокзал"
    TERMINAL = 2, "АС", "Автостанция"
    TICKET_OFFICE = 3, "АК", "Автокасса"
    PLATFORM = 4, "ОП", "Остановочная платформа"


class RouteType(SarmatAttribute):
    """Типы маршрутов."""
    TURNOVER = 1, "turn", "Оборотный"
    CIRCLE = 2, "circle", "Кольцевой"


class JourneyType(SarmatAttribute):
    """Типы рейсов."""
    SUBURBAN = 1, "", "Пригородный"
    LONG_DISTANCE = 2, "", "Междугородный"
    INTER_REGIONAL = 3, "", "Межрегиональный"
    INTERNATIONAL = 4, "", "Международный"


class DurationMonthCalcStrategy(SarmatAttribute):
    """Стратегия расчёта месячной продолжительности."""
    DOWN = 1, "down", "Откат к наименьшему дню месяца"
    MOVE = 2, "move", "Смещение по количеству дней в конце месяца"
    FULL = 3, "max", "Считаем, что во всех месяцах 31 день"


class CalculationType(SarmatAttribute):
    """Виды расчёта стоимости."""
    CONST = 1, "const", "Константное значение"
    RATE = 2, "rate", "Тариф за пройденный километраж"
    STEP = 3, "step", "Зональный тип расчёта стоимости"
    MIXED = 4, "mixed", "Смешанный тип расчёта"


class JourneyClass(SarmatAttribute):
    """Классификация рейсов."""
    BASE = 1, "", "Формирующийся"
    TRANSIT = 2, "", "Транзитный"
    ARRIVING = 3, "", "Прибывающий"


class JourneyState(SarmatAttribute):
    """Состояния рейсов."""
    READY = 0, "", "Активен"
    ARRIVED = 1, "", "Прибыл"
    ON_REGISTRATION = 2, "", "На регистрации"
    DEPARTED = 3, "", "Отправлен"
    CANCELLED = 4, "", "Отменен"
    CLOSED = 5, "", "Закрыт"
    DISRUPTED = 6, "", "Сорван"


class VehicleType(SarmatAttribute):
    """Тип транспортного средства."""
    BUS = 1, "", "Автобус"
    SMALL_BUS = 2, "", "Автобус малой вместимости"
    CAR = 3, "", "Легковой автомобиль"
    TRUCK = 4, "", "Грузовой автомобиль"
    TRAILER = 5, "", "Прицеп"
    SPECIAL = 6, "", "Спецтехника"


class CrewType(SarmatAttribute):
    """Тип участника экипажа."""
    DRIVER = 1, "", "Водитель"
    TRAINEE = 2, "", "Стажер"


class PermitType(SarmatAttribute):
    """Тип путевого листа."""
    BUS_PERMIT = 1, "", "Путевой лист автобуса"
    CAR_PERMIT = 2, "", "Путевой лист легкового автомобиля"
    TRUCK_PERMIT = 3, "", "Путевой лист грузового автомобиля"
    CUSTOM_PERMIT = 4, "", "Заказной путевой лист"


class PlaceKind(SarmatAttribute):
    """Тип места."""
    PASSENGERS_SEAT = 1, "P", "Пассажирское место"
    BAGGAGE = 2, "B", "Багажное место"


class PlaceType(SarmatAttribute):
    """Вид пассажирского места."""
    STANDING = 1, "", "Место для стоящих пассажиров"
    SITTING = 2, "", "Место для сидящих пассажиров"


class PlaceState(SarmatAttribute):
    """Состояние места."""
    FREE = 1, "", "Свободно"
    BOOKED = 2, "", "Забронировано"
    CLOSED = 3, "", "Закрыто"
    SOLD = 4, "", "Продано"
    LOCKED = 5, "", "Заблокировано"
    TRANSFERRED = 6, "", "Произведена пересадка"


MAX_SUBURBAN_ROUTE_LENGTH = 50      # Максимальная длина пригородных маршрутов (в километрах)



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Вычисления с участием Sarmat объектов.
Вычисление метрик.
"""
from sarmat.core.constants import RouteType
from sarmat.core.context.models import RouteMetrics

from .const import MetricsSource


def calculate_route_metrics(source: MetricsSource) -> RouteMetrics:
    """Вычисление метрики для маршрута или рейса.

    Args:
        source: JourneyModel | RouteModel - маршрут или рейс

    Returns: метрика
    """
    last_item = source.structure[-1]

    # Для кольцевых маршрутов последний пункт совпадает с начальным,
    #    поэтому вычислять фактическое имя маршрута не имеет смысла
    if source.route_type == RouteType.CIRCLE:
        route_name = source.name
    else:
        route_name = f"{source.start_point.name} - {last_item.point.name}"

    metrics = RouteMetrics(
        fact_route_name=route_name,
        points_count=len(source.structure),
        total_length=last_item.length_from_last_km,
        spent_time=last_item.travel_time_min,
        move_time=last_item.travel_time_min,
        stop_time=0,
    )

    for item in source.structure[:-1]:
        metrics.total_length += item.length_from_last_km
        metrics.spent_time += (item.travel_time_min + (item.stop_time_min or 0))
        metrics.move_time += item.travel_time_min
        metrics.stop_time += item.stop_time_min or 0

    return metrics



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Модели.
Модели для описания географических объектов.
"""
from dataclasses import dataclass
from typing import Self

from sarmat.core.constants import LocationType, SettlementType

from .sarmat_models import BaseIdModel, BaseModel, CustomAttributesModel


@dataclass
class BaseGeoModel(BaseModel):
    """Модель географического справочника (основные атрибуты)"""

    name: str                           # наименование
    location_type: LocationType         # тип образования
    latin_name: str = ""                # латинское название
    mapping_data: dict | None = None    # данные геолокации
    tags: str = ''                      # теги
    parent: Self | None = None          # родительский объект


@dataclass
class GeoModel(BaseIdModel, CustomAttributesModel, BaseGeoModel):
    """Модель географического справочника"""


@dataclass
class BaseDestinationPointModel(BaseModel):
    """Модель для описания пунктов назначения (основные атрибуты)"""

    name: str                       # наименование
    state: GeoModel                 # территориальное образование
    point_type: SettlementType      # тип поселения


@dataclass
class DestinationPointModel(BaseIdModel, CustomAttributesModel, BaseDestinationPointModel):
    """Модель для описания пунктов назначения"""


@dataclass
class BaseDirectionModel(BaseModel):
    """Модель для описания направления (основные атрибуты)"""

    name: str           # наименование
    cypher: str = ""    # шифр (системное имя)


@dataclass
class DirectionModel(BaseIdModel, CustomAttributesModel, BaseDirectionModel):
    """Модель для описания направления"""


@dataclass
class BaseRoadNameModel(BaseModel):
    """Модель для описания дороги (основные атрибуты)"""

    cypher: str
    name: str = ''


@dataclass
class RoadNameModel(BaseIdModel, CustomAttributesModel, BaseRoadNameModel):
    """Модель для описания дороги"""



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Инструменты для маршрутной сети.
"""
from datetime import (
    UTC,
    date,
    datetime,
    time,
    timedelta,
)
from typing import Generator

from sarmat.core.constants.exception_constants import (
    ErrorClass,
    SarmatWrongValueError,
)
from sarmat.core.constants.sarmat_constants import (
    DurationMonthCalcStrategy,
    JourneyType,
    RouteType,
)
from sarmat.core.context.models import (
    DurationModel,
    JourneyItemModel,
    JourneyModel,
    RouteItemModel,
    RouteModel,
)
from sarmat.core.context.tools.calculators import get_timedelta_from_duration

# Последовательность, возвращаемая генератором для определения последнего пункта в составе маршрута
RouteSeq = tuple[list[RouteItemModel], bool]


def make_journeys_from_route(
    route: RouteModel,
    departure_time: time,
    pauses: list[DurationModel],
    departure_date: date,
    duration_calc_strategy: DurationMonthCalcStrategy,
) -> list[JourneyModel]:
    """Создание рейсов на основе шаблона (маршрута).

    Args:
        route: RouteModel - маршрут
        departure_time: time (UTC) - время отправления для первого рейса
        pauses: list[DurationModel] - список пауз между рейсами
        departure_date: date - расчётная дата отправления (для расчёта продолжительности в месяцы или годы)
        duration_calc_strategy: DurationMonthCalcStrategy - стратегия переноса дней в конце месяца

    Returns: list[JourneyModel] - список построенных маршрутов

    Если маршрут задан как "кольцевой", то точки перелома в составе маршрута не учитываются.
    Будет создан один рейс.
    Выполняется проверка того, что последний пункт в составе маршрута совпадает с пунктом отправления.

    Для оборотных маршрутов:
     - если точка перелома не задана, то будет создан один рейс. Актуально для заказных перевозок.
     - если одна точка перелома указана для последнего пункта в составе маршрута,
       то будут построены два рейса с противоположным составом маршрута.
     - если переломных точек несколько, то будет создано несколько рейсов: от точки до точки.
     - если последний пункт не является переломной точкой, то обратные рейсы не строятся

     Список пауз используется для обозначения перерывов между рейсами.
     Ко времени завершения предыдущего рейса будет добавлена указанная продолжительность,
       чтобы вычислить время отправления следующего рейса.
     Если продолжительность не указана, то время отправления следующего рейса
       будет совпадать со временем завершения предыдущего рейса.

    """
    datetime_counter = datetime.combine(
        departure_date,
        departure_time,
        tzinfo=UTC,
    )
    journey_comment = f"Создан на основе маршрута {route.id} ({route.name})"
    pauses_gen = (i for i in pauses)
    journeys_list = []

    # Создание рейса на основе кольцевого маршрута
    if route.route_type == RouteType.CIRCLE:
        if route.start_point.id != route.structure[-1].point.id:
            raise SarmatWrongValueError(
                err_class=ErrorClass.DATA,
                title="Ошибка создания рейса",
                description="Последний пункт кольцевого маршрута должен соответствовать начальному пункту",
            )
        # Создание рейсов по количеству оборотов маршрута
        first_item = RouteItemModel(
            length_from_last_km=0,
            travel_time_min=0,
            point=route.start_point,
            position=0,
            station=route.departure_station,
        )
        route_items = [first_item] + route.structure
        for _ in range(route.turnovers):
            journey, datetime_counter = create_journey_from_route(
                route,
                datetime_counter,
                route_items,
                journey_comment,
            )
            journeys_list.append(journey)
            datetime_counter = apply_time_delta_to_start_date(
                datetime_counter,
                pauses_gen,
                duration_calc_strategy,
            )
    else:
        need_to_build_turnovers = False
        # общее количество рейсов в прямом и обратном направлении
        # будет равно количеству рейсов в одном круге, помноженному на количество оборотов
        for _ in range(route.turnovers):
            turnover_items = []
            first_item = RouteItemModel(
                length_from_last_km=0,
                travel_time_min=0,
                point=route.start_point,
                position=0,
                station=route.departure_station,
            )
            # обход состава маршрута в прямом направлении
            for sub_structure, is_final in route_structure_gen(route.structure):
                route_items = [first_item] + sub_structure
                journey, datetime_counter = create_journey_from_route(
                    route,
                    datetime_counter,
                    route_items,
                    journey_comment,
                )
                journeys_list.append(journey)
                turnover_items.append(route_items[::-1])
                datetime_counter = apply_time_delta_to_start_date(
                    datetime_counter,
                    pauses_gen,
                    duration_calc_strategy,
                )
                first_item = RouteItemModel(
                    length_from_last_km=0,
                    travel_time_min=0,
                    point=sub_structure[-1].point,
                    position=0,
                    station=sub_structure[-1].station,
                )
                if is_final:
                    need_to_build_turnovers = sub_structure[-1].is_breaking_point

            if need_to_build_turnovers:
                # обход состава маршрута в обратном направлении
                for sub_structure in turnover_items[::-1]:
                    journey, datetime_counter = create_journey_from_route(
                        route,
                        datetime_counter,
                        sub_structure,
                        journey_comment,
                    )
                    journeys_list.append(journey)
                    datetime_counter = apply_time_delta_to_start_date(
                        datetime_counter,
                        pauses_gen,
                        duration_calc_strategy,
                    )

    return journeys_list


def create_journey_from_route(
    route: RouteModel,
    start_time: datetime,
    items: list[RouteItemModel],
    comment: str = "",
) -> tuple[JourneyModel, datetime]:
    """Вспомогательная функция по созданию рейса из маршрута.

    Args:
        route: RouteModel - маршрут
        start_time: datetime - дата и время отправления рейса
        items: list[RouteItemModel] - состав маршрута
        comment: str - комментарий к рейсу

    Returns: JourneyModel, datetime - модель рейса и дата с приращением времени выполнения рейса

    """

    if not items:
        raise SarmatWrongValueError(
            err_class=ErrorClass.DATA,
            title="Ошибка построения рейса",
            description="Пункт отправления и состав маршрута должны быть определены",
        )
    start_point, *structure = items
    if not structure:
        raise SarmatWrongValueError(
            err_class=ErrorClass.DATA,
            title="Ошибка построения рейса",
            description="Состав маршрута должны быть определён",
        )

    current_time = start_time
    new_journey = JourneyModel(
        route_type=route.route_type,
        name=f"{start_point.point.name} - {structure[-1].point.name}",
        start_point=start_point.point,
        departure_station=start_point.station,
        direction=route.direction,
        comments=comment,
        journey_type=JourneyType.LONG_DISTANCE,
        departure_time=current_time.time(),
        structure=[],
    )

    for idx, item in enumerate(structure):
        current_time += timedelta(minutes=item.travel_time_min)
        stop_time = item.stop_time_min or 0

        new_journey.structure.append(
            JourneyItemModel(
                departure_time=(current_time + timedelta(minutes=stop_time)).time(),
                arrive_time=current_time.time(),
                length_from_last_km=item.length_from_last_km,
                travel_time_min=item.travel_time_min,
                point=item.point,
                position=idx,
                station=item.station,
                stop_time_min=item.stop_time_min,
            )
        )
        current_time += timedelta(minutes=stop_time)

    # В последнем пункте время отправления не нужно
    new_journey.structure[-1].departure_time = None
    return new_journey, current_time


def route_structure_gen(route_structure: list[RouteItemModel]) -> Generator[RouteSeq, None, None]:
    """Генератор для разбиения состава маршрута по точкам перелома.

    Args:
        route_structure: list[RouteItemModel] - элементы из состава маршрута

    Returns: генератор, разбирающий состав маршрута по точкам перелома.

    """
    sub_route = []
    item_last_idx = len(route_structure) - 1

    for idx, item in enumerate(route_structure):
        sub_route.append(item)
        if item.is_breaking_point:
            yield sub_route, idx == item_last_idx
            sub_route = []

    if sub_route:
        yield sub_route, True


def apply_time_delta_to_start_date(
    departure_datetime: datetime,
    pauses_gen: Generator[DurationModel, None, None],
    duration_calc_strategy: DurationMonthCalcStrategy,
) -> datetime:
    """Приращение по времени для определения времени отправления для следующего рейса.

    Args:
        departure_datetime: datetime - дата и время отправления
        pauses_gen: Generator - генератор продолжительностей
        duration_calc_strategy: DurationMonthCalcStrategy - стратегия расчёта переноса дней

    Returns: datetime - дата и время с приращением времени из очередного элемента продолжительности

    """
    try:
        pause = next(pauses_gen)
    except StopIteration:
        return departure_datetime

    return departure_datetime + get_timedelta_from_duration(
        duration=pause,
        calculation_date=departure_datetime.date(),
        calc_strategy=duration_calc_strategy,
        only_active=False,
    )



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Вычисления с участием Sarmat объектов.
Вычисление стоимости проезда.
"""
from abc import ABC, abstractmethod
from typing import Any, Union

from dataclasses import asdict, dataclass

from sarmat.core.constants import SarmatWrongValueError, ErrorClass
from sarmat.core.constants.sarmat_constants import CalculationType
from sarmat.core.context.models import (JourneyModel, RouteModel)

RouteSimilarity = Union[JourneyModel, RouteModel]


@dataclass
class PriceCalculation:
    """Структура для расчёта стоимости."""

    @property
    def as_dict(self):
        return asdict(self)


@dataclass
class PriceCalculationConstant(PriceCalculation):
    """Структура для простого расчёта стоимости."""

    values_map: list[list[float]]    # значения
    calc_type: CalculationType = CalculationType.CONST


@dataclass
class PriceCalculationRate(PriceCalculation):
    """Структура для расчёта стоимости за пройденный километраж."""

    rate: float     # значение тарифа
    calc_type: CalculationType = CalculationType.RATE


@dataclass
class PriceCalculationStep(PriceCalculation):
    """Структура для расчёта стоимости по зонам прохождения."""

    steps: dict[int, float]     # маппинг стоимости на расстояние
    calc_type: CalculationType = CalculationType.STEP


class PriceCalculator(ABC):
    """Калькулятор стоимости проезда."""

    def __init__(self):
        self._calc_data = None

    @abstractmethod
    def get_price_calculation_structure(self, args: dict[str, Any]) -> PriceCalculation:
        """Получение структуры для расчёта стоимости проезда.

        Args:
            args: сериализованные данные

        Returns: структура со значениями для расчёта стоимости
        """

    @abstractmethod
    def calculate_price(
        self,
        route: RouteSimilarity,
        departure_index: int,
        destination_index: int,
    ) -> float:
        """Базовый метод вычисления стоимости проезда.

        Args:
            route: рейс или маршрут
            departure_index: индекс пункта отправления
            destination_index: индекс пункта назначения

        Returns: высчитанная стоимость
        """
        if departure_index >= destination_index:
            raise SarmatWrongValueError(
                err_class=ErrorClass.DATA,
                title="Ошибка расчёта стоимости",
                description="Пункт отправления должен быть раньше пункта назначения",
            )

        route_len = len(route.structure)
        # NOTE: учитывается начальный пункт, который не присутствует в списке
        if max(departure_index, destination_index) > route_len:
            raise SarmatWrongValueError(
                err_class=ErrorClass.DATA,
                title="Ошибка расчёта стоимости",
                description="Указанный индекс превышает длительность маршрута",
            )

        if self._calc_data is None:
            raise SarmatWrongValueError(
                err_class=ErrorClass.DATA,
                title="Ошибка расчёта стоимости",
                description="Не указаны параметры расчёта стоимости",
            )
        return 0.0


class PriceCalculatorForConstant(PriceCalculator):
    """Калькулятор для расчёта стоимости по обычному способу."""

    def get_price_calculation_structure(self, args: dict[str, Any]) -> PriceCalculation:
        """Получение структуры для расчёта стоимости проезда.

        Args:
            args: сериализованные данные

        Returns: структура со значениями для расчёта стоимости
        """
        return PriceCalculationConstant(**args)

    def set_calculation_data(self, calc_data: PriceCalculationConstant) -> None:
        """Указание расчётных данных."""
        self._calc_data = calc_data

    def calculate_price(
        self,
        route: RouteSimilarity,
        departure_index: int,
        destination_index: int,
    ) -> float:
        """Базовый метод вычисления стоимости проезда.

        Args:
            route: рейс или маршрут
            departure_index: индекс пункта отправления
            destination_index: индекс пункта назначения

        Returns: высчитанная стоимость
        """
        super().calculate_price(route, departure_index, destination_index)
        return self._calc_data.values_map[departure_index][destination_index]


class PriceCalculatorForRate(PriceCalculator):
    """Калькулятор для расчёта стоимости по тарифу за пройденное расстояние."""

    def get_price_calculation_structure(self, args: dict[str, Any]) -> PriceCalculation:
        """Получение структуры для расчёта стоимости проезда.

        Args:
            args: сериализованные данные

        Returns: структура со значениями для расчёта стоимости
        """
        return PriceCalculationRate(**args)

    def set_calculation_data(self, calc_data: PriceCalculationRate) -> None:
        """Указание расчётных данных."""
        self._calc_data = calc_data

    def calculate_price(
        self,
        route: RouteSimilarity,
        departure_index: int,
        destination_index: int,
    ) -> float:
        """Базовый метод вычисления стоимости проезда.

        Args:
            route: рейс или маршрут
            departure_index: индекс пункта отправления
            destination_index: индекс пункта назначения

        Returns: высчитанная стоимость
        """
        super().calculate_price(route, departure_index, destination_index)
        sub_route = slice(departure_index, destination_index)
        amount = 0.0
        for item in route.structure[sub_route]:
            amount += self._calc_data.rate * item.length_from_last_km

        return amount


class PriceCalculatorForStep(PriceCalculator):
    """Калькулятор для расчёта стоимости по зонному тарифу."""

    def get_price_calculation_structure(self, args: dict[str, Any]) -> PriceCalculation:
        """Получение структуры для расчёта стоимости проезда.

        Args:
            args: сериализованные данные

        Returns: структура со значениями для расчёта стоимости
        """
        return PriceCalculationStep(**args)

    def set_calculation_data(self, calc_data: PriceCalculationStep) -> None:
        """Указание расчётных данных."""
        self._calc_data = calc_data

    def calculate_price(
            self,
            route: RouteSimilarity,
            departure_index: int,
            destination_index: int,
    ) -> float:
        """Базовый метод вычисления стоимости проезда.

        Args:
            route: рейс или маршрут
            departure_index: индекс пункта отправления
            destination_index: индекс пункта назначения

        Returns: высчитанная стоимость
        """
        super().calculate_price(route, departure_index, destination_index)
        sub_route = slice(departure_index, destination_index)
        traveled_distance = sum(
            [item.length_from_last_km for item in route.structure[sub_route]]
        )
        steps = [(traveled_distance, 0)] + [
            (distance, value) for distance, value in self._calc_data.steps.items()
        ]
        # NOTE: сортировка по обратным величинам стоимости
        #       позволяет при совпадении пройденного расстояния с табличным
        #       отодвинуть вставленный элемент вправо
        #       и не получить нулевую стоимость проезда
        length, values = zip(*sorted(steps, key=lambda x: (x[0], -x[1])))
        idx = length.index(traveled_distance)

        return values[idx+1] if idx else max(values[idx], values[idx+1])



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Модели.
Подвижной состав.
"""
from dataclasses import dataclass
from datetime import date
from typing import List

from sarmat.core.constants import CrewType, PermitType, VehicleType

from .sarmat_models import PersonModel, BaseIdModel, BaseUidModel, BaseModel, CustomAttributesModel


@dataclass
class BaseVehicleModel(BaseModel):
    """Подвижной состав (основные атрибуты)"""

    vehicle_type: VehicleType   # тип транспортного средства
    vehicle_name: str           # марка транспортного средства
    state_number: str           # гос. номер
    seats: int                  # количество мест для посадки
    stand: int = 0              # количество мест стоя
    capacity: int = 0           # вместимость багажного отделения


@dataclass
class VehicleModel(BaseIdModel, CustomAttributesModel, BaseVehicleModel):
    """Подвижной состав"""


@dataclass
class BaseCrewModel(BaseModel):
    """Экипаж (основные атрибуты)"""

    crew_type: CrewType     # тип члена экипажа
    is_main: bool = True    # признак главного члена экипажа


@dataclass
class CrewModel(BaseIdModel, CustomAttributesModel, BaseCrewModel, PersonModel):
    """Экипаж"""


@dataclass
class BasePermitModel(BaseModel):
    """Путевой лист (основные атрибуты)"""

    number: str                     # номер путевого листа
    permit_type: PermitType         # тип путевого листа
    depart_date: date               # дата выезда
    crew: List[CrewModel]           # экипаж
    vehicle: List[VehicleModel]     # подвижной состав


@dataclass
class PermitModel(BaseUidModel, CustomAttributesModel, BasePermitModel):
    """Путевой лист"""



"""
Sarmat.
Ядро пакета.
Описание ошибок.
Типы ошибок, коды, описание.
"""
from collections import namedtuple
from dataclasses import dataclass
from enum import Enum


SarmatErrorType = namedtuple("SarmatErrorType", "code title")
SarmatErrorClass = namedtuple("SarmatErrorClass", "cypher description")


class MessageType(SarmatErrorType, Enum):
    """Типы сообщений."""
    INFORMATION = 0, "Информация"
    QUESTION = 1, "Вопрос"
    WARNING = 2, "Внимание"
    ERROR = 3, "Ошибка"
    CRITICAL = 4, "Критическая ошибка"


class ErrorClass(SarmatErrorClass, Enum):
    """Классификация ошибки."""

    UNKNOWN = "", ""
    SYSTEM = "S", "Системная ошибка"
    DATA = "D", "Ошибка данных"
    OPERATION = "O", "Ошибка выполнения операции"


@dataclass
class SarmatException(Exception):
    """Класс ошибки в формате Sarmat."""
    err_class: ErrorClass = ErrorClass.UNKNOWN
    err_type: MessageType = MessageType.ERROR
    title: str = ""
    description: str = ""


class SarmatExpectedAttributeError(SarmatException):
    """Ошибка возникает при отсутствии атрибута в объекте."""
    err_class = ErrorClass.DATA
    title = "Отсутствует атрибут"


class SarmatNotFilledAttribute(SarmatException):
    """Ошибка возникает при обнаружении незаполненного атрибута."""
    err_class = ErrorClass.DATA
    title = "Не указано значение"


class SarmatWrongTypeAttribute(SarmatException):
    """Атрибут содержит значение неверного типа."""
    err_class = ErrorClass.DATA
    title = "Невертный тип данных"


class SarmatWrongValueError(SarmatException):
    """Атрибут содержит неверное значение."""
    err_class = ErrorClass.DATA
    title = "Неверное значение данных"


class SarmatWrongOperationError(SarmatException):
    """Невозможно выполнить операцию."""
    err_class = ErrorClass.OPERATION
    title = "Невозможная операция"



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Модели.
Модели для построения маршрутной сети.
"""
from dataclasses import dataclass
from datetime import time, date

from sarmat.core.constants import JourneyType, RoadType, RouteType, StationType

from .geo_models import DirectionModel, DestinationPointModel, RoadNameModel
from .sarmat_models import BaseIdModel, BaseModel, CustomAttributesModel, DurationModel


@dataclass
class BaseStationModel(BaseModel):
    """Станции (пункты посадки-высадки пассажиров) (основные атрибуты)."""

    station_type: StationType           # тип станции
    name: str                           # наименование
    point: DestinationPointModel        # ближайший населенный пункт
    address: str = ''                   # почтовый адрес


@dataclass
class StationModel(BaseIdModel, CustomAttributesModel, BaseStationModel):
    """Станции (пункты посадки-высадки пассажиров)."""


@dataclass
class BaseRoadModel(BaseModel):
    """Дороги (основные атрибуты)."""

    start_point: DestinationPointModel          # начало дороги
    end_point: DestinationPointModel            # конец дороги
    direct_travel_time_min: int                 # время прохождения в прямом направлении
    reverse_travel_time_min: int                # время прохождения в обратном направлении
    direct_len_km: float                        # расстояние в прямом направлении
    reverse_len_km: float                       # расстояние в обратном направлении
    road_type: RoadType                         # тип дорожного покрытия
    road_name: RoadNameModel | None = None      # классификация дороги


@dataclass
class RoadModel(BaseIdModel, CustomAttributesModel, BaseRoadModel):
    """Дороги."""


@dataclass
class BaseRouteItemModel(BaseModel):
    """Состав маршрута (основные атрибуты)."""

    length_from_last_km: float              # расстояние от предыдущего пункта
    travel_time_min: int                    # время движения от предыдущего пункта в минутах
    point: DestinationPointModel            # точка прохождения маршрута
    position: int = 1                       # порядок следования
    station: StationModel | None = None     # станция
    stop_time_min: int | None = None        # время стоянки в минутах
    road: RoadModel | None = None           # дорога


@dataclass
class RouteItemModel(BaseIdModel, CustomAttributesModel, BaseRouteItemModel):
    """Состав маршрута."""

    is_breaking_point: bool = False     # точка перелома маршрута (начало следующего рейса)


@dataclass
class CommonRouteModel(BaseModel):
    """Общие атрибуты для маршрута и рейса."""

    route_type: RouteType                           # тип маршрута
    name: str                                       # наименование
    start_point: DestinationPointModel              # точка отправления
    departure_station: StationModel | None = None   # станция отправления
    direction: list[DirectionModel] | None = None   # направления
    comments: str | None = None                     # комментарий к маршруту
    number: int | None = None                       # номер маршрута
    literal: str = ''                               # литера
    is_active: bool = True                          # признак активности маршрута


@dataclass
class BaseRouteModel(BaseModel):
    """Описание маршрута (основные атрибуты)."""

    structure: list[RouteItemModel]     # состав маршрута


@dataclass
class RouteModel(BaseIdModel, CustomAttributesModel, CommonRouteModel, BaseRouteModel):
    """Описание маршрута."""

    turnovers: int = 1      # количество оборотов по маршруту


@dataclass
class JourneyItemModel(BaseIdModel, CustomAttributesModel, BaseRouteItemModel):
    """Состав рейса."""

    departure_time: time | None = None      # время отправления из пункта
    arrive_time: time | None = None         # время прибытия в пункт


@dataclass
class BaseJourneyModel(BaseModel):
    """Атрибуты рейса (основные атрибуты)."""

    journey_type: JourneyType           # тип рейса
    departure_time: time                # время отправления
    structure: list[JourneyItemModel]   # структура рейса


@dataclass
class JourneyModel(BaseIdModel, CustomAttributesModel, CommonRouteModel, BaseJourneyModel):
    """Атрибуты рейса"""

    is_chartered: bool = False          # признак заказного рейса
    need_control: bool = False          # признак именной продажи и мониторинга
    season_begin: date | None = None    # начало сезона
    season_end: date | None = None      # окончание сезона


@dataclass
class RouteMetrics:
    """Описание метрики маршрута или рейса."""

    fact_route_name: str    # фактическое название маршрута
    points_count: int       # количество пунктов в составе
    total_length: float     # общая протяжённость маршрута
    spent_time: int         # затраченное время в минутах
    move_time: int          # время в движении
    stop_time: int          # время стоянки на промежуточных пунктах


@dataclass
class BaseJourneyBunchItemModel(BaseModel):
    """Атрибуты элемента из связки рейсов (основные атрибуты)"""

    journey: JourneyModel           # рейс
    stop_duration: DurationModel    # время простоя
    position: int                   # номер в последовательности


@dataclass
class JourneyBunchItemModel(BaseIdModel, CustomAttributesModel, BaseJourneyBunchItemModel):
    """Атрибуты элемента из связки рейсов"""


@dataclass
class BaseJourneyBunchModel(BaseModel):
    """Атрибуты связки рейсов (основные атрибуты)"""

    journeys: list[JourneyBunchItemModel]   # элементы связки


@dataclass
class JourneyBunchModel(BaseIdModel, CustomAttributesModel, BaseJourneyBunchModel):
    """Атрибуты связки рейсов"""

    name: str | None = None                 # наименование связки



"""
Sarmat.

Используемые константы.

Форматы значений.
"""
DATE_FORMAT = '%d.%m.%Y'
TIME_FORMAT = '%H:%M'
FULL_TIME_FORMAT = '%H:%M:%S'
DATETIME_FORMAT = f'{DATE_FORMAT} {TIME_FORMAT}'
FULL_DATETIME_FORMAT = f'{DATE_FORMAT} {FULL_TIME_FORMAT}'



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Модели.
"""
__all__ = (
    "DestinationPointModel", "DirectionModel", "GeoModel", "RoadNameModel", "BaseModel", "StationModel", "RoadModel",
    "RouteModel", "RouteItemModel", "JourneyModel", "JourneyBunchModel", "JourneyBunchItemModel", "IntervalModel",
    "JourneyProgressModel", "CrewModel", "JourneyScheduleModel", "PermitModel", "VehicleModel", "PersonModel",
    "DurationModel", "DurationItemModel", "IntervalItemModel", "JourneyIntervalModel", "JourneyItemModel",
    "RouteMetrics",
)

from .dispatcher_models import (
    JourneyIntervalModel,
    JourneyProgressModel,
    JourneyScheduleModel,
)
from .geo_models import (
    DestinationPointModel,
    DirectionModel,
    GeoModel,
    RoadNameModel,
)
from .sarmat_models import (
    BaseModel,
    DurationItemModel,
    DurationModel,
    IntervalItemModel,
    IntervalModel,
    PersonModel,
)
from .traffic_management_models import (
    JourneyBunchItemModel,
    JourneyBunchModel,
    JourneyItemModel,
    JourneyModel,
    RoadModel,
    RouteItemModel,
    RouteMetrics,
    RouteModel,
    StationModel,
)
from .vehicle_models import (
    CrewModel,
    PermitModel,
    VehicleModel,
)



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Модели.
Диспетчерские объекты.
"""
from dataclasses import dataclass
from datetime import date, datetime
from typing import List, Optional

from sarmat.core.constants import JourneyClass, JourneyState

from .sarmat_models import BaseIdModel, BaseModel, BaseUidModel, CustomAttributesModel, IntervalModel
from .traffic_management_models import DestinationPointModel, JourneyModel, StationModel
from .vehicle_models import PermitModel


@dataclass
class BaseJourneyIntervalModel(BaseModel):
    """График выполнения рейсов (основные атрибуты)"""

    journey: JourneyModel       # рейс
    start_date: date            # дата начала
    interval: IntervalModel     # интервал движения


@dataclass
class JourneyIntervalModel(BaseIdModel, CustomAttributesModel, BaseJourneyIntervalModel):
    """График выполнения рейсов"""


@dataclass
class BaseJourneyProgressModel(BaseModel):
    """Атрибуты рейсовой ведомости (основные атрибуты класса)"""

    depart_date: date           # дата отправления в рейс
    journey: JourneyModel       # рейс
    permit: PermitModel         # номер путевого листа


@dataclass
class JourneyProgressModel(BaseUidModel, CustomAttributesModel, BaseJourneyProgressModel):
    """Атрибуты рейсовой ведомости"""


@dataclass
class BaseJourneyScheduleModel(BaseModel):
    """Процесс прохождения рейса по автоматизированным точкам (основные атрибуты)"""

    journey_progress: JourneyProgressModel                       # рейсовая ведомость
    journey_class: JourneyClass                                  # классификация рейса в данном пункте
    station: Optional[StationModel]                              # станция
    point: Optional[DestinationPointModel]                       # точка прохождения маршрута
    state: JourneyState                                          # состояние рейса
    plan_arrive: Optional[datetime] = None                       # плановое время прибытия
    fact_arrive: Optional[datetime] = None                       # фактическое время прибытия
    plan_depart: Optional[datetime] = None                       # плановое время отправления
    fact_depart: Optional[datetime] = None                       # фактическое время отправления
    platform: str = ''                                           # платформа
    comment: str = ''                                            # комментарий к текущему пункту
    last_items: Optional[List['JourneyScheduleModel']] = None    # оставшиеся активные пункты прохождения рейса


@dataclass
class JourneyScheduleModel(BaseUidModel, CustomAttributesModel, BaseJourneyScheduleModel):
    """Процесс прохождения рейса по автоматизированным точкам"""



"""
Sarmat.
Ядро пакета.
Описание бизнес логики.
Вспомогательные инструменты.
Вычисления с участием Sarmat объектов.
Константы общего назначения.
"""
from typing import Union

from sarmat.core.context.models import (
    JourneyModel,
    RouteModel,
)

max_month_len = 31
months_in_year = 12

MetricsSource = Union[JourneyModel, RouteModel]
